;;; -*- auto-recompile: t -*-

;;; csound-lsp.el --- Csound composition in Emacs Lisp

;; Keywords: csound

;; This file is not part of GNU Emacs.
;; 
;; csound-lsp.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; csound-lsp.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Author: Stphane Rollandin <hepta@zogotounga.net>
;;---------

;;; Commentary:
;;-------------

;;    full documentation here:
;;    http://www.zogotounga.net/comp/csoundx.html

;;
;;; Installation:
;;                  
;; ==========================================================
;; this file should be installed through the csound-x package
;; ==========================================================
;;

;; last modified May 10, 2007

;;; Code:

(require 'csound-csd)
(require 'csound-orc)

(defgroup csound-lsp nil
  "Emacs Lisp for Csound composition"
  :group 'csound-x
  :prefix "csl-")

(defcustom csl-sr 44100
  "Default sample rate"
  :type 'integer
  :group 'csound-lsp)

(defcustom csl-kr 4410
  "Default control rate"
  :type 'integer
  :group 'csound-lsp)

(defcustom csl-nchnls 1
  "Default number of ouput channels"
  :type 'integer
  :group 'csound-lsp)

(defcustom csl-variable-output-opcodes '("soundin" "diskin" "diskin2")
  "opcodes known to have a variable number of outputs"
  :type '(repeat string)
  :group 'csound-lsp)


;; ===================================================================
;;                   formatting utilities
;; ===================================================================

(defmacro with-csd-template (&rest body)
  `(progn
     (insert "<CsoundSynthesizer>\n")
     ,@body
     (insert "\n</CsoundSynthesizer>\n")))

(defmacro with-csl-options-template (&rest body)
  `(progn
     (insert "\n<CsOptions>\n")
     ,@body
     (insert "\n</CsOptions>\n")))

(defmacro with-csl-score-template (&rest body)
  `(progn
     (insert "\n<CsScore>\n")
     ,@body
     (insert "\n</CsScore>\n")))

(defmacro with-csl-orchestra-template (&rest body)
  `(progn
     (setq csl-varname-counter 1000) 
     (insert (format "\n<CsInstruments>
sr     = %s
kr     = %s
ksmps  = %s
nchnls = %s

" csl-sr csl-kr (/ csl-sr csl-kr) csl-nchnls))
     ,@body
     (insert "\n</CsInstruments>\n")))

(defvar csl-varname-counter 1000)
(make-variable-buffer-local 'csl-varname-counter)

(defun csl-new-varname (&optional opcode)
  (format "%s%s" (if opcode (substring (concat opcode "vv") 0 3)
                   "var")
          (incf csl-varname-counter)))

;TEST (let ((csl-varname-counter 0)) (csl-new-varname "linen")) => "lin1"
;TEST (let ((csl-varname-counter 0)) (csl-new-varname "oscili")) => "osc1"
;TEST (let ((csl-varname-counter 0)) (csl-new-varname "in")) => "inv1"


;; ===================================================================
;;               orchestra code transcription engine
;; ===================================================================

(defun csl-transcribe-to-code (sexp)
  "Transcribe the specification SEXP into orchestra code lines 
in the usual Csound syntax

SEXP valid forms are as follow:

\(OPCODE [:global] [:xrate] ARGS)
where OPCODE is a Csound opcode and ARGS its arguments;
any argument can itself be a specification SEXP.
:xrate is either :arate, :krate or :irate

\(OPCODE \(:output OUTVARS) ARGS)
in which case the opcode output variables are direcly given
ex: \(oscili \(:output aosc) p4 1/p3 p7)

\(:code STRING) for litteral inclusion of STRING as a code line

\(::MACRO BODY) where BODY structure depends on ::MACRO
this is documented in the associated function csl-expand::MACRO
for example if ::MACRO is ::add or ::line, the corresponding
functions are `csl-expand::add' and `csl-expand::line'
"
  ;; ::macros expansion
  (when (and (keywordp (first sexp))
             (string-match "^::" (symbol-name (first sexp))))
    (setq sexp 
          (funcall 
           (symbol-function (intern (format "csl-expand%s" (first sexp))))
           sexp)))
  ;;
  (if (listp (first sexp)) ;; jump to sub-level if required
      (loop for s in sexp
            for o = (csl-transcribe-to-code s)
            append (cadr o) into code-lines
            finally return (list (car o) code-lines))
    ;;
    (let* ((opcode (symbol-name (first sexp)))
           output opargs code-lines globalp)
      ;;
      ;; litteral inclusion
      (if (eq :code (first sexp))
          (setq code-lines (list (apply 'concat (cdr sexp))))
        ;;
        ;; regular parsing
        (dolist (arg (cdr sexp))
          (cond
           ;; single keyword 
           ((keywordp arg)
            (cond
             ((eq arg :global) (setq globalp t))
             ((memq arg '(:arate :krate :irate))
              (setq output (csl-output-for opcode arg nil globalp)))
             (t (error "unrecognized keyword: %s" arg))))
           ;; list
           ((listp arg)
            (case (first arg)
              ;; output spec (:output k1 a2 ...)
              (:output (setq output (mapcar 'symbol-name (cdr arg))))
              ;; else another sexp to expand
              (t (let ((rec (csl-transcribe-to-code arg)))
		   (setq opargs (append opargs (car rec))
			 code-lines (append code-lines (cadr rec)))))))
           ;; anything else is considered an opcode argument
           (t (add-to-list 'opargs (format "%S" arg) t))))
        (setq output (or output
                         (csl-output-for opcode nil opargs globalp)))
        (add-to-list 'code-lines
                     (if (member opcode csdoc-functions)
                         (format "%s = %s" 
                                 (car output)
                                 (if (null (cdr opargs))
                                     (concat opcode "(" (first opargs) ")")
                                   (mapconcat 'identity opargs opcode)))
                       (format "%s %s %s"
                               (mapconcat 'identity output ", ") 
                               opcode
			       ;; to do: handle possible mismatch in input format:	      
			       (mapconcat 'identity
					  (butlast opargs 
						   (- (length opargs)
						      (length (csdoc-opcode :in opcode))))
					  ", "))) 
                     t))
      (list output code-lines))))

(defun csl-debug-ttc (sexp)
  (let ((csl-varname-counter 0)) 
    (csl-transcribe-to-code sexp)))

;TEST (csl-debug-ttc '(out (oscili :arate (linen :krate p4 p6 p3 p7) p5 p8)))) => '(nil ("klin2 linen p4, p6, p3, p7" "aosc1 oscili klin2, p5, p8" " out aosc1"))

(defun csl-output-for (opcode &optional rate opargs globalp)
  (cond
   ((member opcode csl-variable-output-opcodes)
    (error "unspecified outputs for %s" opcode))
   ((not (member opcode csdoc-opcodes))
    (if (yes-or-no-p (format "Unknown opcode %s. Proceed with warning ? " opcode))
        (list (format ";WARNING: variables missing (unknown opcode %s)\n" opcode))
      (error "Unknown opcode %s" opcode)))
   ((member opcode csdoc-functions)
    (list (concat
           (when globalp "g")
           (if (member opcode '("a" "k" "i"))
               opcode
             (if (null (cdr opargs))
                 (substring (car opargs) 0 1)
               (let ((spec (mapcar (lambda (v) (elt (remove ?g v) 0)) opargs)))
                 (cond ((member ?a spec) "a")
                       ((member ?k spec) "k")
                       (t "i")))))
           (csl-new-varname))))
   ((member opcode csdoc-0-opcodes)
    ())
   (t (let ((spec (substitute ?k ?x ;; should the default be ?a or ?k ?
                              (car (split-string (csdoc-opcode :spec opcode)))))
            (output))
        (dotimes (o (length spec))
          (add-to-list 'output
                       (concat (when globalp "g")
                               (string (if rate
                                           (elt (symbol-name rate) 1)
                                         (elt spec o)))
                               (csl-new-varname opcode))
                       t))
        output))))

;TEST (elt (car (csl-output-for "linen" :arate)) 0) => 97
;TEST (elt (car (csl-output-for "linen" :krate)) 0) => 107
;TEST (elt (car (csl-output-for "linen")) 0) => 107
;TEST (elt (car (csl-output-for "+" nil '("ahoui" "kbof"))) 0) => 97
;TEST (elt (car (csl-output-for "+" nil '("gahoui" "kbof"))) 0) => 97
;TEST (elt (car (csl-output-for "+" nil '("khoui" "kjhu" "kbof"))) 0) => 107
;TEST (elt (car (csl-output-for "+" nil '("khoui" "kjhu" "abof"))) 0) => 97
;TEST (elt (car (csl-output-for "a" nil '("khoui" "kjhu" "ibof"))) 0) => 97


;; ===================================================================
;;                 orchestra code transcription macros
;; ===================================================================

(defun csl-expand::cond (arg)
  "Usage: \(::cond \(TEST SEXP [SEXP ...]) [\(TEST SEXP [SEXP ...]) ...])
"
  (loop for s in (nthcdr 1 arg)
        with exps = nil
        if (null exps)
        collect `(:code ,(format "if %s then" (car s))) into exps
        else
        collect `(:code ,(format "elseif %s then" (car s))) into exps
        append `(,@(cdr s)) into exps
        finally return (append exps '((:code "endif")))))

;TEST (csl-debug-ttc '(::cond ((p4 < 20) (oscili (:output aosc1) p4 1/p3 p7) (oscili (:output aosc) p4 1/p3 p7)) ((5 == 2) (oscili (:output kbof) 1 2 3)))))  => '(nil ("if (p4 < 20) then" "aosc1 oscili p4, 1/p3, p7" "aosc oscili p4, 1/p3, p7" "elseif (5 == 2) then" "kbof oscili 1, 2, 3" "endif"))

(defun csl-expand::if (arg)
  "Usage: \(::if TEST SEXP [SEXP ...])
"
  (append `((:code ,(format "if %s then" (second arg))))
          (nthcdr 2 arg)
          `((:code "endif"))))

;TEST (csl-debug-ttc '((::if (p4 < 2000) (oscili (:output aosc1) p4 1/p3 p7) (oscili (:output aosc) p4 1/p3 p7)) (out aosc))) => '(nil ("if (p4 < 2000) then" "aosc1 oscili p4, 1/p3, p7" "aosc oscili p4, 1/p3, p7" "endif" " out aosc"))

(defun csl-expand::loop (arg)
  "Usage: \(::loop LOOP-CLAUSES SEXP-TEMPLATE)
where SEXP-TEMPLATE is a quoted or backquoted sexp.
...see `loop' for valid loop clauses
"
  (eval `(loop ,@(subseq arg 1 -1)
               collect ,@(last arg))))

;TEST (csl-debug-ttc '(::loop for i upto 5 `(init (:output ,(intern (format "ga%s" i))) 0))) => '(("ga5") ("ga0 init 0" "ga1 init 0" "ga2 init 0" "ga3 init 0" "ga4 init 0" "ga5 init 0"))

(defun csl-expand::add (arg)
  "Usage: \(::add LOOP-CLAUSES SEXP-TEMPLATE)
where SEXP-TEMPLATE is a quoted or backquoted sexp 
...see `loop' for valid loop clauses

example:

   \(::add for amp in '\(p4 2*p4 p4/2)
          `\(oscili :krate ,amp 1/p3 p7))

is equivalent to:

   \(+ \(oscili :krate p4 1/p3 p7)
      \(oscili :krate 2*p4 1/p3 p7)
      \(oscili :krate p4/2 1/p3 p7))
"
  (append (list '+)
          (eval `(loop ,@(subseq arg 1 -1)
                       collect ,@(last arg)))))

;TEST (csl-debug-ttc '(::add for amp in '(p4 p4/2) for freq in `(1/p3 ,(/ 6 2)) `(oscili :krate ,amp ,freq p7))) => '(("kvar3") ("kosc1 oscili p4, 1/p3, p7" "kosc2 oscili p4/2, 3, p7" "kvar3 = kosc1+kosc2"))

(defun csl-expand::line (arg)
  "Usage: \(::line LOOP-CLAUSES SEXP-TEMPLATE)
where SEXP-TEMPLATE is a quoted or backquoted  sexp in which 
one element \(or subelement) must be a vector.
...see `loop' for valid loop clauses

example:

   \(::line for amp in '\(p4 2*p4 p4/2)
           `\(oscili ,amp [1/p3] p7))

is equivalent to:

   \(oscili p4/2 \(oscili 2*p4 \(oscili p4 1/p3 p7) p7) p7))
"
  (let* ((templ (car (last arg)))
         (seed nil)
         (plug-form (loop for s in (cadr templ)
                          until (vectorp s)
                          finally (setq seed (car (append s nil)))
                          finally return s))     
         (form nil))
    (eval `(loop ,@(subseq arg 1 -1)
                 do (setq form (eval (subst (or form seed) plug-form templ)))
                 finally return form))))

;TEST (csl-debug-ttc '(::line for freq in `(1/p3 ,(* 2 .1)) `(oscili :krate [p4] ,freq p7))) => '(("kosc1") ("kosc2 oscili p4, 1/p3, p7" "kosc1 oscili kosc2, 0.2, p7"))

'(examples 
  ::add
  (csl-edit-composition
   (csound-composition
    :instr 1
    '(out (::add for amp in '(p4 2*p4 p4/2)
                 for freq in `(1/p3 ,(/ 1 10.0) ,(+ .01 0.05))
                 `(oscili :krate ,amp ,freq p7)))))
  ::line
  (csl-edit-composition
           (csound-composition
            :instr 10
            '(:code "; this is a test")
            '(out (::line for freq in `(1/p3 ,(/ 1 10.0) ,(+ .01 0.05))
                          `(oscili :krate [p4] ,freq p7)))))

  (csl-edit-composition
           (csound-composition
            :instr 10
            '((:code "; this is a test")
              (out (::line for freq in `(1/p3 ,(/ 1 10.0) ,(+ .01 0.05))
                          `(oscili :krate [p4] ,freq p7))))))
  ::if 
  (csl-edit-composition
           (csound-composition
            :instr 1
            '((::if (p4 < 2000)
                    (oscili (:output aosc1) p4 1/p3 p7)
                    (oscili (:output aosc) p4 1/p3 p7))
              (out aosc))))
  ::cond 
  (csl-edit-composition
           (csound-composition
            :instr 1
            '((::cond ((p4 < 2000)
                       (oscili (:output aosc1) p4 1/p3 p7)
                       (oscili (:output aosc) p4 1/p3 p7))
                      ((5 == 2)
                       (oscili (:output kbof) 1 2 3)))
              (out aosc))))
  ::loop
  (csl-edit-composition
           (csound-composition
            :globals
            '(::loop for i upto 5
                     `(init (:output ,(intern (format "ga%s" i))) 0))))
end examples)


;; ===================================================================
;;                   the 'csound-composition macro
;; ===================================================================

'(example 
  (csl-edit-composition
;  (csl-play-composition  
   (csound-composition
    :options "-dummy --no joke"
    :sr (* 4 12 1000)
;    :ksmps 12
    :globals '(::loop for i upto 5
                      `(= (:output ,(intern (format "ga%s" i))) 0))
    :ftables '(1 :sin 8192)
    :opcode "dummy,i,ii"
    '(out 0)
    :instr 1
    '(out (oscili :arate (oscili :krate p4 1/p3 p7) p5 p6))
    :score
    (insert "f 2 0 513 5 1 12 1024 500 1" ?\n)
    (i 1 0 4 8000 440 1 2)
    (loop for h in '(2200 600 215 1852 990)
          for d in '(2 2.5 2 1.5 4)
          do (+i :p5 h :p3 d))))
)

(defmacro csl-nconc (var list)
  ;; using nconc in csound-composition leads to weird errors...
  `(dolist (el ,list)
     (add-to-list ',var el t))) 

(defmacro csound-composition (&rest args)
  "Main macro for csound composition
Return a lambda form which when evaluated inserts at point a full CSD.

ARGS structure is \(KEY BODY [KEY BODY] ...) where KEY is a keyword
while BODY does not include a keyword at top-level. a same KEY can appear 
any number of time; no KEY is mandatory.

In the following HEAD refers to the first form in BODY and TAIL
refers to the list of all other forms, nil if BODY is a single form

Supported KEYs are:

:sr, :kr:, :ksmps, :nchnls 
  set the corresponding global to HEAD. this is not mandatory.
  \(defaut values come from variables `csl-sr', `csl-kr' and `csl-nchnls')

:globals
  insert initialisation statements in orchestra, before instruments and 
opcode definitions. 
  BODY forms are orchestra code specs. see `csl-transcribe-to-code'

:ftables
  use all forms in BODY as ftable specifications. see `csl-insert-ftable'

:instr 
  insert an instrument definition in the orchestra
  HEAD is inst number, TAIL are orchestra code specs.

:opcode 
  insert an opcode definition in the orchestra
  HEAD is opcode spec \(a string), TAIL are orchestra code specs.

:orchestra
  evaluate BODY within the orchestra section of the CSD, among instruments
and opcodes definitions. BODY forms is plain lisp code.

:score 
  evaluate BODY within the score section of the CSD, similar to :orchestra

:i-stream 
  same as :score but BODY is wrapped in `with-pfields-stream'

:options
 set the CSD options section \(i.e. csound command line flags)
 BODY forms are converted to string and concatenated 
"
  (let (options
        (cscompo-orc-globals '(void))
        (cscompo-orc-code '(void))
        (cscompo-sco-header '(void))
        (cscompo-sco-statements '(void))
        key body head tail)
    (dolist (chunk 
             (loop for form in args
                   for fplace upfrom 1
                   if (keywordp form)
                   collect `(,form ,@(loop for f in (nthcdr fplace args)
                                           until (keywordp f) 
                                           collect f))))
      (setq key (car chunk)
            body (cdr chunk)
            head (cadr chunk)
            tail (cddr chunk))
      (case key
        (:ftables     
	 (dolist (fspec body)
	   (csl-nconc cscompo-sco-header `((apply 'csl-insert-ftable ,fspec)))))
        (:ksmps
         (csl-nconc cscompo-orc-code `((csound-set-ksmps ,head))))
        (:kr
         (csl-nconc cscompo-orc-code `((csound-set-kr ,head))))
        (:sr
         (csl-nconc cscompo-orc-code `((csound-set-sr ,head))))
        (:nchnls
         (csl-nconc cscompo-orc-code `((csound-set-nchnls ,head))))
        (:instr
         (csl-nconc cscompo-orc-code `((csl-insert-instrument ,head ,@tail))))
        (:opcode 
	 (csl-nconc cscompo-orc-code `((csl-insert-opcode ,head ,@tail))))
        (:orchestra
	 (csl-nconc cscompo-orc-code body))
        (:score
	 (csl-nconc cscompo-sco-statements body))
        (:i-stream       
	 (csl-nconc cscompo-sco-statements `((with-pfields-stream nil ,@body))))
        (:options 
	 (setq options (mapconcat (lambda (f) (format "%s" f)) body " ")))
        (:globals        
           (dolist (code body)
             (csl-nconc cscompo-orc-globals `((csl-insert-code ,code)))))))
    `(lambda ()
       (atomic-change-group
	 (with-csd-template
	  (when ,options
	    (with-csl-options-template
	     (insert ,options)))
	  (with-csl-orchestra-template
	   ,@(cdr cscompo-orc-globals)
	   ,@(cdr cscompo-orc-code))
	  (with-csl-score-template
	   ,@(cdr cscompo-sco-header)
	   ,@(cdr cscompo-sco-statements)))))))

(defun csl-insert-code (quoted-sexp)
  (let ((code (csl-transcribe-to-code quoted-sexp)))
    (dolist (line (cadr code))
      (insert line "\n"))
    (car code)))

(defun csl-insert-instrument (name &rest body)
  (insert (format "\ninstr %s\n" name))
  (dolist (code body)
    (csl-insert-code code))
  (insert "endin\n"))

(defun csl-insert-opcode (name &rest body)
  (insert (concat "\nopcode " name "\n"))
  (dolist (code body)
    (csl-insert-code code))
  (insert "endop\n"))

(put 'csl-insert-instrument 'lisp-indent-function 1)

(defun csl-play-composition (composition)
  (with-csound-to-DAC (funcall composition)))

(defun csl-render-composition (composition wavfile)
  (with-csound-to-WAV wavfile (funcall composition)))

(defcustom csl-composition-buffer-name "Csound composition"
  "Name for buffers created for editing a csound composition"
  :type 'string
  :group 'csound-lsp)

(defun csl-edit-composition (composition)
  (switch-to-buffer (generate-new-buffer csl-composition-buffer-name))
  (funcall composition)
  (goto-char (point-min))
  (csound-csd-mode))


;; ===================================================================
;;                   f-tables utilities
;; ===================================================================

(defcustom csl-tables '(:sin "f %s 0 %s 10 1")
  "Templates for GEN tables
First replacement slot is for table number, second for table size"
  :type 'plist
  :group 'csound-lsp)

(defun csl-insert-ftable (num type size)
  (insert (format (plist-get csl-tables type) num (expt 2 (round (log size 2)))) ?\n))


;; ===================================================================
;;                   testing
;; ===================================================================

(defvar csl-example-composition 
  (csound-composition
   :instr 1108
   '(out (oscili :arate (linen :krate p4 p6 p3 p7) p5 p8))
   :instr 1109
   '(out (* (:output aout) 
            (linen (:output kline) p4 p6 p3 p7)
            (+ (oscili (:output aosc1) 1 p5 p9)
               (oscili (:output aosc2) 1 p8 p9))))
   :ftables '(1 :sin 8192)
   :score
   (i 1108 0 2 15000 400 .3 .3 1)
   (i 1108 3 2 15000 10 .3 .3 1)
   (i 1109 6 2 25000 400 .3 .3 10 1)
   ;; (s)
   (insert "s\n")
   (i 1108 1 2 15000 400 .3 .3 1)
   (i 1108 4 2 15000 170 .3 .3 1)
   (i 1109 7 2 25000 400 .3 .3 170 1)  
   ;; (s)
   (insert "s\n")
   (i 1108 1 2 15000 400 .3 .3 1)
   (i 1108 4 2 15000 385 .3 .3 1)
   (i 1109 7 2 25000 400 .3 .3 385 1)))

(defun csl-test-dac ()
  (interactive)
  (csl-play-composition csl-example-composition))

(defun csl-test-wav ()
  (interactive)
  (csl-render-composition csl-example-composition "test.wav"))

;; === this is it.
(provide 'csound-lsp)

;; csound-lsp.el ends here

